package sde.to.model;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.apache.log4j.Logger;

import sde.to.util.CAssertion;
import sde.to.util.CUtility;

/**
 * 時系列データをあたえるクラス
 * 
 * （時間、値）
 * 
 * 
 * @author ohno
 *
 */
public class CTS {
	

	// log4jの設定
	final Logger logger = Logger.getLogger(CTS.class);
	
	/**
	 * 時系列データ　値　
	 */
	private Map<Double, Double> _data;
	
	/**
	 * 最大値
	 */
	private double max;
	
	/**
	 * 最大値（時系列）
	 */
	private Map<Double,Double> _maxdata;
	/**
	 * コンストラクタ
	 */
	public CTS(){
		//　時系列データの初期設定
		this._data = new TreeMap<Double, Double>();
		// 最大時系列データの初期設定
		this._maxdata = new TreeMap<Double, Double>();
		// 最大値の初期設定
		this.max = Double.MIN_VALUE;
	}
	
	/**
	 * 時系列データの取得
	 */
	public Map<Double,Double> getData(){
		// 時系列データの取得
		return this._data;
	}
	/**
	 * 時系列データの取得
	 * @param a　取得開始時刻
	 * @param b　取得終了時刻
	 * @return key(t)が[a,b]に入る時系列データをかえす。
	 */
	public CTS getSubCTS(double a,double b){
		// 返す値の用意
		CTS rtn = new CTS();
		
		//　イテレータの取得
		Iterator<Double> its = this._data.keySet().iterator();
		// 鍵の取得
		while(its.hasNext()){
			double key = its.next();
			if( key>= a && key <= b){
				rtn.put(key, this._data.get(key));
			}
		}
		// 返却
		return rtn;
	}
	/**
	 * 最大値の取得
	 */
	public Map<Double,Double> getMaxData(){
		
		// 時系列データの取得
		return this._maxdata;
	}
	
	/**
	 * データ挿入
	 * @param t 時間
	 * @param v 値
	 * 
	 */
	public void put(double t,double v){
		
		this._data.put(t, v);
		
		// 最大値の設定
		this.max = Math.max(v, this.max);
		
		// 最大時系列データへの設定
		this._maxdata.put(t, this.max);
	}
	
	/**
	 * 配列（Y値　値）の取得
	 */
	public Object[] getY(){
		return this._data.values().toArray();
	}
	
	/**
	 * 配列（X値 時間）の取得
	 */
	public Object[] getX(){
		return this._data.keySet().toArray();
	}
	
	/**
	 * データサイズの返還
	 */
	public int size(){
		return this._data.size();
	}
	
	/**
	 * データの返還
	 * @param x key
	 * @return データ
	 */
	public double get(double x){
		return this._data.get(x);
	}
	
	/**
	 * 最大値の返還
	 * @return 最大値
	 */
	public double getMax(){
		return this.max;
	}
	
	/**
	 * 時刻　t までの最大値をかえす。
	 * 
	 * t >=0.
	 * 
	 * @return 最大値
	 */
	public double getMax(double t){
		
		// t が鍵として含まれるか？
		if(this._maxdata.containsKey(t)== true)
			return this._maxdata.get(t);
		
		// 含まれない場合
		double key = this.getRound(t);
		
		return this._maxdata.get(key);
			
		
	}
	
	/**
	 * 時刻 t までの　epsilon 以下 alpha を越えるのkey数を返す。
	 * 
	 * @param t
	 * @param epsilon
	 * @param alpha
	 * @return measure
	 * 
	 */
	public long measure(double t,double epsilon,double alpha){
		//　鍵イテレータの取得
		Iterator<Double> it = this._data.keySet().iterator();
		// カウンターの初期化
		long cnt = 0;
		while(it.hasNext()){
			// 鍵の取得
			double key = it.next();
			// epsilon 以下、時刻が　t　以下
			if(key <= t && this._data.get(key) <= epsilon && this._data.get(key) > alpha)cnt++;
			
		}
		return cnt;
		
	}
	/**
	 * FirstStoppingTimeの返還
	 * 
	 *       *----------y1(bvalue)
	 *       |  x2(key)
	 * a-----|---|--------
	 *       x1  |
	 *     (bkey)|
	 *           *------y2(value)
	 *           
	 *            =((y1-a)x2 +x1(a-y2))/(y1-y2)
	 *            
	 *  return 値　=( (bvalue-a)key + bkey(a-value))/(bvalue-value)
	 *  
	 * 値が設定されない場合は最大値(Double.MAX)をかえす。
	 * @param a 
	 * @return FirstStoppingTime
	 */
	public double getFirstStoppingTime(double a){
		//
		Iterator<Double> it = this._data.keySet().iterator();
		
		// フラグ　-1 (a > value) , 1 (a < value) , 0 初期値
		short flg = 0;
		
		//鍵の一時保管領域
		double bkey = 0.;
		//値の一時保管領域
		double bvalue = 0.;
		
		// データ（鍵）繰り返し
		while(it.hasNext()){
			// 一つとりだす。
			double key = it.next();
			// 値をとりだす。
			double value = this._data.get(key);
			
			// aとの比較
			if(a == value)return key;
			
			// 返す値の設定
			double x = ( (bvalue-a)*key + bkey*(a-value))/(bvalue-value);
			// 今回はじめて a以下となる。
			if(a > value && flg == 1)return x;
			
			// 今回はじめて　a以上となる。
			if(a < value && flg == -1)return x;
			
			//フラグを設定する。
			if(a > value) {
				// フラグの設定
				flg = -1;
				// 値の一時保管
				bvalue = value;
				// 鍵の一時保管
				bkey = key;
			}
			//
			if(a < value ) {
				flg = 1;
				// 値の一時保管
				bvalue = value;
				// 鍵の一時保管
				bkey = key;
			}
		}
		// firstStoppingTimeが設定されない場合は、-1をかえす。
		return Double.MAX_VALUE;
	}
	
	/**
	 * 近い鍵のデータを返す。
	 * @param x key
	 * @return xの最近の鍵に対応するデータ
	 * 
	 */
	public double getRound(double x){
		//
		Iterator<Double> it = this._data.keySet().iterator();
		// 初期値設定
		double d = Double.MAX_VALUE;
		// key
		double key = 0.;
		// データ（鍵）繰り返し
		while(it.hasNext()){
			// 一つとりだす。
			double one = it.next();
			// xとの比較
			if(d > Math.abs(x-one)){
				// x に近い鍵をkeyに設定する。
				d = Math.abs(x-one);
				// x に近い鍵をkeyに設定する。
				key = one;
				
			}
			
		}
		// 値をかえす。　ない場合は、例外が発生する。
		return this.get(key);
	}
	

	/**
	 * 印字用のデータのヘッダーをつくる。
	 * <-----n----><-----n---->
	 * 1              2
	 * <-5->         <-5->
	 */
	public String getHeaderString(int n){
		String rtn = "";
		n=(int)(0.9*(double)n);// space の幅　と　数値のフォント幅の違いを吸収（てきとう）
		Iterator<Double> its  = this._data.keySet().iterator();
		while(its.hasNext()){
			double key = its.next();
			try {
				rtn += CUtility.cvDouble(key,1.,5);
				rtn += CUtility.getSp(n - rtn.length()%n);
			} catch (CAssertion e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
		}
		return rtn;
	}
	
	
	
	/**
	 * 印字用のデータの返還
	 * <-----n----><-----n---->
	 * **.*****      **.******
	 *    <-5->         <-5->
	 *  
	 * @param n 印字データ間隔
	 * @param m 改行
	 * @param isIdx 格行の先頭にインデクスをつけるか
	 */
	public String getString(int n,int m,boolean isIdx){
		String rtn = "";
		int i = 0;
		
		if(isIdx == true)rtn+=""+i+CUtility.getSp(n - (""+i).length()%n);
		Iterator<Double> its  = this._data.keySet().iterator();
		while(its.hasNext()){
			double key = its.next();
			try {
				String one =""+CUtility.cvDouble(this._data.get(key),1.,5);
				rtn += one;
				rtn += CUtility.getSp(n - one.length()%n);
				if(++i % m == 0){
					if(isIdx == false)
					rtn +="\n";
					else
						rtn +="\n"+i+CUtility.getSp(n - (""+i).length()%n);
					
				}
			} catch (CAssertion e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
		}
		return rtn;
	}
	
	/**
	 * 印字用のデータの返還(指数返還する）
	 * 
	 * @param n 印字データ間隔
	 */
	public String getStringExp(int n){
		String rtn = "";
		Iterator<Double> its  = this._data.keySet().iterator();
		while(its.hasNext()){
			double key = its.next();
			try {
				rtn += CUtility.cvDouble(Math.exp(this._data.get(key)),1.,5);
				rtn += CUtility.getSp(n - rtn.length()%n);
			} catch (CAssertion e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
		}
		return rtn;
	}
	
	/**
	 * 印字用のデータの返還(対数返還する）
	 * 
	 * @param n 印字データ間隔
	 */
	public String getStringLog(int n){
		String rtn = "";
		Iterator<Double> its  = this._data.keySet().iterator();
		while(its.hasNext()){
			double key = its.next();
			try {
				rtn += CUtility.cvDouble(Math.log(this._data.get(key)),1.,5);
				rtn += CUtility.getSp(n - rtn.length()%n);
			} catch (CAssertion e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
		}
		return rtn;
	}
	
	/**
	 * 指数変換
	 * 
	 * @return 
	 * 		値をexp(.) してかえす。
	 */
	public CTS getExp() {
		CTS rtn = new CTS();
		Iterator<Double> its = this._data.keySet().iterator();
		while (its.hasNext()) {
			double key = its.next();
			rtn.put(key, Math.exp(this.get(key)));

		}
		return rtn;
	}
	
	/**
	 * 零集合(値=0の場合）を与える。
	 * 指定した値をとる推定key値のリストをかえす。
	 * @param value
	 * @return 0-list(CTS)
	 */
	public CTS getZeroSet(double value){
		// 返却値の用意
		CTS rlist = new CTS();
		// 値の格納域 初期値（最大値をいれる）
		double one = Double.MAX_VALUE;
		//
		double two = Double.MAX_VALUE;
		// 
		double keytwo = 0;
		
		// イテレーション
		Iterator<Double> its = this._data.keySet().iterator();
		//　値を２つとり、零点を生成するかどうかを決める。
		while (its.hasNext()) {
			// oneにコピー（最初は、MAX値が設定）
			one = two;
			// keyをコピー
			double keyone = keytwo;
			// keyの取得
			keytwo = its.next();
			// twoに値を取得
			two = this._data.get(keytwo);
			//二つの値 one , two からZeroセットに追加するか。
			if(check(one,two,value)== true)
				// 追加するKEY値を結果に入れる。
				rlist.put(mk(one,two,keyone,keytwo,value),value);
		}
		
		// 返却
		return rlist;
	}
	
	/**
	 * getZeroSetに使用される関数
	 * @param one
	 * @param two
	 * @param value
	 * @return 追加する場合にtrue
	 */
	private boolean check(double one,double two,double value){
		// oneが未設定の場合
		if(one == Double .MAX_VALUE)
			{
			if(two == value) return true;
			else return false;
			}
		//
		if( two == value)
			return true;
		//
		if(one < value && two > value)
			return true;
		
		if(one > value && two < value)
			return true;
		
		return false;
		
		
	}
	
	/**
	 * getZeroSetに使用される関数
	 * 
	 * .(x1,y1)=one
	 *  .
	 * --x-----------value
	 *    .
	 *     .(x2,y2)=two
	 * 
	 *  x= (x1*(y2-value)+x2*(value-y1))/(y2-y1)
	 * 
	 * 
	 * @param y1
	 * @param y2
	 * @param x1
	 * @param x2
	 * @param value
	 * @return 追加する値
	 */
	private double mk(double y1,double y2,double x1,double x2,
			double value){
		// oneが未設定の場合
		if(y1 == Double .MAX_VALUE)
			return x2;
		
		if(y2 == value)
			return x2;
		
		// 次の条件のみ。
		//if((y1 < value && y2 > value) || (y1 > value && y2 < value))
			return (x1*(y2-value)+x2*(value-y1))/(y2-y1);
	
		
		
	}

	/**
	 * locat time をかえす。
	 * 
	 * @param value
	 * @param eps
	 * 
	 * @return localtime
	 * 
	 */
	public double getLocalTime(double value,double eps) {
		
		// 値の格納域 初期値（最大値をいれる）
		double y1 = Double.MAX_VALUE;
		//
		double y2 = Double.MAX_VALUE;
		//
		double x2 = 0;
		// 戻し値
		double localtime = 0.;

		// イテレーション
		Iterator<Double> its = this._data.keySet().iterator();
		// 　値を２つとり、零点を生成するかどうかを決める。
		while (its.hasNext()) {
			// oneにコピー（最初は、MAX値が設定）
			y1 = y2;
			// keyをコピー
			double x1 = x2;
			// keyの取得
			x2 = its.next();
			// twoに値を取得
			y2 = this._data.get(x2);
		
			if(y1 !=  Double.MAX_VALUE ){
				localtime += this.localT(y1, y2, x1, x2, value, eps);
			}
		}
		
		// localtimeを戻す。
		return localtime;

	}
	
	/**
	 * mkLocalTimeに使用される関数
	 *    
	 *    .(x1,y1)
	 * ----|---------------- + eps
	 *     |.
	 * ----|--------------- value
	 *     |  .
	 * ----|--|.------------ - eps
	 *     |  | .(x2,y2)
	 *     |  | 
	 *   ->|  |<- return 値
	 *          
	 */
	private double localT(double y1,double y2 ,double x1,double x2,double value ,double eps){
		// eps > 0.
		assert (eps > 0.);
		// 交点がなく０をかえす場合
		// min(y1,y2) > value + eps or
		// max(y1,y2) < value - eps
		// 交点がなく　x2 - x1をかえす場合
		// min(y1,y2) > value - eps and 
		// max(y1,y2) < value + eps
		// 交点が一点　value + epsのみ
		// max(y1,y2) >= value + eps and
		// min(y1,y2) > value - eps
		// 交点が1点　value - epsのみ
		// max(y1,y2) < value + eps and
		// min(y1,y2) <= value -eps
		// 交点が2点
		// max(y1,y2) >= value + eps and
		// min(y1,y2) <= vlaue - eps
		//
		double max = Math.max(y1, y2);
		//
		double min = Math.min(y1, y2);
		//
		double vp = value + eps;
		//
		double vm = value - eps;
		// 交点の計算
		double xp = this.crossX(y1, y2, x1, x2, vp);
		//
		double xm = this.crossX(y1, y2, x1, x2, vm);
		//中間点の設定
		double x = x1;
		//
		if(vp > y2 && vm < y2)x = x2;
		
		
		//
		double X = Math.abs(xp-xm);
		
		// 
		double rtn = 0.;
		
		//
		//  .
		//---xp.----------
		//    .
		//-----xm---------
		//     .
		//
		if(max > vp){
			if(min > vp){
				rtn =  0.;
			} else if(min == vp){
				rtn =  0.;
			} else if(min > vm){
				rtn = Math.abs( x - xp);
			} else if(min == vm){
				rtn =  X;
			} else if(min < vm){
				rtn =  X;
			}
			
		} else if(max == vp){
			if(min > vp){
				assert(false);
			} else if(min == vp){
				rtn =  0.;
			} else if(min > vm){
				rtn = Math.abs( x - xp);
			} else if(min == vm){
				rtn =  X;
			} else if(min < vm){
				rtn =  X;
			}
			
		} else if(max > vm){
			if(min > vp){
				assert(false);
			} else if(min == vp){
				assert(false);
			} else if(min > vm){
				rtn =  X;
			} else if(min == vm){
				rtn =  X;
			} else if(min < vm){
				rtn =  Math.abs(xm - x);
			}
			
		} else if(max == vm){
			if(min > vp){
				assert(false);
			} else if(min == vp){
				assert(false);
			} else if(min > vm){
				assert(false);
			} else if(min == vm){
				rtn = 0.;
			} else if(min < vm){
				rtn = 0.;
			}
			
		} else if(max < vm){
			if(min > vp){
				assert(false);
			} else if(min == vp){
				assert(false);
			} else if(min > vm){
				assert(false);
			} else if(min == vm){
				assert(false);
			} else if(min < vm){
				rtn = 0.;
				
			}
			
		}
		
		// 整合性のチェック
		if(rtn <0.){
			logger.debug("max["+max+"],min["+min+"],X["+X+"],x["+x+"]");
			logger.debug("x1["+x1+"],y1["+y1+"],x2["+x2+"],y2["+y2+"],vp["+vp+"],vm["+vm+"],rtn["+rtn+"]");
			logger.debug("xp["+xp+"],xm["+xm+"]");
		}
		assert(rtn >= 0.);
		
		// 値をかえす。
		return rtn;
		
		
		
	}
	

	/**
	 * mkLocalTimeに使用される関数
	 * 2013.3.23
	 * .(x1,y1)=one
	 *  .                  
	 * --x---------------vlue(x,y)
	 *    .                
	 *     .(x2,y2)=two
	 * 
	 * y given -> return x
	 * 
	 * 交点がない場合は、
	 * y < min(y1,y2)  and if y1 < y2 return x1 else return x2
	 * y > max(y1,y2)  and if y1 < y2 return x2 else return x1
	 * 
	 * @param y1
	 * @param y2
	 * @param x1
	 * @param x2
	 * @param y
	 * @return x
	 */
	private double crossX(double y1,double y2,double x1,double x2,
			double y){
		// x1 < x2を仮定
		assert(x1 < x2);
		
		// 交点がある場合
		if( (y-y1)*(y-y2) <= 0.){
			if(y == y1) return x1;
			if(y == y2) return x2;
			
			double a = (y - y2)/(y1 - y);
			
			return (x2 + a*x1)/(a + 1);
			//  a+1 == 0 --> y1 == y2 
		}
		
		// 交点がない場合
		if(y < Math.min(y1,y2) ){
			if( y1 < y2) return x1;
			else return x2;
		}
		//
		if(y > Math.max(y1,y2)){
			if(y1 < y2 ) return x2;
			else return x1;
		}
		
		// ダミー
		assert(false);
		//
	return 0.;
	
	
		
		
	}
	
	
	
	/**
	 * しきい値以下の個数をかえす。
	 * @param x しきい値
	 * @return #{ y(k) =< x }
	 */
	public long A(double x){
		long n = 0;
		Iterator<Double> its  = this._data.keySet().iterator();
		while(its.hasNext()){
			double key = its.next();
			try {
				if(this._data.get(key) <= x)n++;
			} catch (CAssertion e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
			
			
			return n;
			
	}
	/**
	 * しきい値以下の鍵がK以下のデータ数をかえす。
	 * 
	 * @param K keyの上限
	 * @param x しきい値
	 * @return #{ y(k) =< x , k =< K}
	 */
	public long A(double K,double x){
		long n = 0;
		Iterator<Double> its  = this._data.keySet().iterator();
		while(its.hasNext()){
			double key = its.next();
			try {
				if(this._data.get(key) <= x && key <= K)n++;
			} catch (CAssertion e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
			
			
			return n;
			
	}
	
	/**
	 * トータル（valueの総和）を返す。
	 */
	public double total(){
		// イテレータの準備
		Iterator<Double> its  = this._data.keySet().iterator();
		//　トータル値初期化
		double sum = 0.;
		while(its.hasNext()){
			// 鍵の取得
			double key = its.next();
			// 値の取得
			double value = this._data.get(key);
			// 和
			sum+=value;
		}
		// トータル値の返却
		return sum;
	}
	/**
	 * 平均値（value2+value1)/2*(key2-key1)の総和）を返す。
	 * keyにDouble.MAX_VALUEがないこを前提としている。
	 */
	public double mean(){
		// イテレータの準備
		Iterator<Double> its  = this._data.keySet().iterator();
		//　平均値初期化
		double mean = 0.;
		//　鍵（x値）の初期化
		double key1 = Double.MAX_VALUE;
		//
		double value1 = Double.MAX_VALUE;
		while(its.hasNext()){
			
			// 鍵の取得
			double key2 = its.next();
			// 値の取得
			double value2 = this._data.get(key2);
			//
			if(key1 != Double.MAX_VALUE){
				mean += (value2 + value1)/2.*(key2 - key1);
			}
			//
			key1 = key2;
			//
			value1 = value2;
			
			
		}
		// 平均値の返却
		return mean;
	}
	
	/**
	 * 最初に登録された(key,value)を印字する。
	 */
	public void dumpFirst(){
		Iterator<Double> its  = this._data.keySet().iterator();
		while(its.hasNext()){
			double key = its.next();
			double value = this._data.get(key);
			logger.debug("["+key+","+value+"]");
			break;
		}
	}
	
	/**
	 * ダンプ
	 */
	public void dump(){
		Iterator<Double> its  = this._data.keySet().iterator();
		while(its.hasNext()){
			double key = its.next();
			double value = this._data.get(key);
			logger.debug("["+key+","+value+"]");
		}
	}
	

}
