package sde.to.util;

import org.apache.log4j.Logger;


/**
 * 正規分布 乱数発生器
 * 
 * @author ohno
 *
 */
public class CNormalRG extends CRandomGenerator {
	
	// log4jの設定
	final Logger logger = Logger.getLogger(CNormalRG.class);
	
	/**
	 * 平均
	 */
	private double myu;
	
	/**
	 * 分散
	 */
	private double siguma2;
	
	
	/**
	 * コンストラクタ
	 * 
	 * @param myu 平均
	 * @param siguma2 分散
	 * @throws MCMCAssertion
	 */
	public CNormalRG(double myu , double siguma2) throws CAssertion{
		
		if(siguma2 <= 0. )
			throw new CAssertion("CNormalRG: siguma2<= 0.");
		
		this.myu = myu;
		
		this.siguma2 = siguma2;
		
		
	}
	

	/**
	 * コンストラクタ
	 * 
	 * @throws MCMCAssertion
	 */
	public CNormalRG() {
		
		this.myu = 0.;
		//2012.11.07 0.->1. bugs nouuse
		this.siguma2 = 1.;
		
	}
	
	
	
	/**
	 * 
	 * 乱数生成
	 * 
	 * @return 乱数値
	 */
	public double next(){
		
		double siguma = Math.sqrt(this.siguma2);
		
		return this.nextGaussian()*siguma + this.myu;
	}
	
	/**
	 * 
	 * 乱数生成
	 * コンストラクタで設定した平均値　分散値とは
	 * 関係なく、パラメータであたえられた平均と分散値で
	 * 乱数を発生する。
	 * 
	 * @param myu
	 * @param siguma2
	 * @return 乱数値
	 */
	public double next(double myu, double siguma2){
		
		double siguma = Math.sqrt(siguma2);
		
		return this.nextGaussian()*siguma + myu;
	}
	
	
	/**
	 * シードの設定
	 * @param seed シード
	 */
	public void setSeed(int seed){
		super.setSead(seed);
	}
	

}
