package sde.to.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.GregorianCalendar;

import org.apache.commons.math3.analysis.UnivariateFunction;
import org.apache.commons.math3.linear.RealVector;
import org.apache.log4j.Logger;

import sde.to.model.CTS;
import tai.utility.TOChartLine;
import tai.utility.TOChartPlot;


/**
 * ユーティティ 2012.03.06
 * 
 * @author ohno
 * 
 */
public class CUtility {

	// log4jの設定
	static final Logger logger = Logger.getLogger(CUtility.class);
	
	/**
	 * PNGPATH
	 */

	static String PNGPATH = "./PNG/";
	
	/**
	 * RealVector print
	 * 
	 */
	public static void ptRealVector(RealVector v){
		for (int i = 0; i<v.getDimension();i++)
		System.out.println("v["+i+"]="+v.getEntry(i));
		
	}
	/**
	 * Object[] to double[]
	 * 
	 * @param D
	 *            Objectの配列
	 * @return 実数の配列
	 */
	public static double[] Otod(Object[] D) {
		double[] rtn = new double[D.length];
		for (int i = 0; i < D.length; i++)
			rtn[i] = (Double) D[i];
		return rtn;
	}

	/**
	 * Integer[] to int[]
	 * 
	 * @param I
	 *            Objectの配列
	 * @return 整数の配列
	 */
	public static int[] Otoi(Object[] I) {
		int[] rtn = new int[I.length];
		for (int i = 0; i < I.length; i++)
			rtn[i] = (Integer) I[i];
		return rtn;
	}

	/**
	 * int[] to double[]
	 * 
	 * @param I
	 *            整数の配列
	 * @return double型の配列
	 */
	public static double[] itod(int[] I) {
		double[] rtn = new double[I.length];
		for (int i = 0; i < I.length; i++)
			rtn[i] = I[i];
		return rtn;
	}

	/**
	 * 今日の日付を返す。 2010.12.22
	 * 
	 * @return 日付型文字列
	 */
	public static String getToDay() {
		Calendar cal = new GregorianCalendar();
		final SimpleDateFormat fmt = new SimpleDateFormat("yyyy-MM-dd-H");
		return fmt.format(cal.getTime());
	}
	
	/**
	 * 実配列内のある値以上の個数
	 * 
	 */
	public static int getPositiveNum(double[][] v ,  int idx,double threshold){
		
		int rtn = 0;
		
		
		
		
		for( int i = 0; i < v.length ; i++){
			
			if(v[i][idx] >= threshold )rtn++;
		}
			
	return rtn;
		
	}

	/**
	 * 時系列データの描画 ファイルー＞CTS
	 * データ
	 * id,val
	 * または
	 * id val
	 * 
	 * @param fpath ファイルパス
	 * @throws IOException
	 * @throws NumberFormatException
	 */
	static public CTS fread(String fpath) throws  IOException {
		CTS res = new CTS();

		// ...ファイルを表すFileクラスを生成。
		File file = new File(fpath);

		try {
			// ファイル入力
			FileReader reader = new FileReader(file);
			// ファイルリーダ
			BufferedReader buff = new BufferedReader(reader);
			// 一行バッファ
			String oneline = null;

		double key=0.,val=0.;
			
			// 一行とりだし
			while ((oneline = buff.readLine()) != null) {
				// u v value
				String[] wards = oneline.split("[, ]");
				try{
				key = Double.parseDouble(wards[0].trim());
				val = Double.parseDouble(wards[1].trim());
				
				}catch( NumberFormatException e){
					logger.debug(e);
	
				}

				res.put(key, val);

			}
		} catch (FileNotFoundException e) {
			// ...事前にチェック済み
			logger.error(e.getMessage());
			assert (false);
		}

		return res;
	}
	/**
	 * 時系列データの描画 ファイルー＞CTS
	 * データ
	 * xxxx=val
	 * 
	 * @param fpath ファイルパス
	 * @throws IOException
	 * @throws NumberFormatException
	 */
	static public CTS fread2(String fpath) throws  IOException {
		CTS res = new CTS();

		// ...ファイルを表すFileクラスを生成。
		File file = new File(fpath);

		// area 計算
		double area = 0.;
		
		try {
			// ファイル入力
			FileReader reader = new FileReader(file);
			// ファイルリーダ
			BufferedReader buff = new BufferedReader(reader);
			// 一行バッファ
			String oneline = null;

			double keyb = 0.;
		double key=0.,val=0.;
			
			// 一行とりだし
			while ((oneline = buff.readLine()) != null) {
				// u v value
				String[] wards = oneline.split("[\\[\\]=]");
				for(int i = 0; i<wards.length;i++){
					logger.debug(wards[i]);
				}
				try{
				String dummy  = wards[0].trim();
				String dummy1  = wards[1].trim();
				String dummy2  = wards[2].trim();
				
				key = Double.parseDouble(wards[3].trim());
				String dummy3 =wards[4].trim();
				val = Double.parseDouble(wards[5].trim());
				
				}catch( NumberFormatException e){
					logger.debug(e);
	
				}

				res.put(key, val);
				
				area += (key-keyb)*val;
				
				keyb = key;

			}
		} catch (FileNotFoundException e) {
			// ...事前にチェック済み
			logger.error(e.getMessage());
			assert (false);
		}

		logger.debug("area["+area+"]");
		
		return res;
	}
	

	/**
	 * 時系列データの描画
	 * 
	 * @param res
	 *            時系列データ（描画の対象）
	 * @param title
	 *            タイトル
	 */
	public static void draw(CTS res, String title) {
		TOChartLine line = new TOChartLine();
		int series_count = 1;
		double[][] value_x = new double[series_count][];
		value_x[0] = new double[res.size()];
		// System.arraycopy(CUtility.itod(CUtility.Otoi(res.getX())), 0,
		// value_x[0], 0, res.size());
		System.arraycopy(CUtility.Otod(res.getX()), 0, value_x[0], 0,
				res.size());
		double[][] value_y = new double[series_count][];
		value_y[0] = new double[res.size()];
		System.arraycopy(CUtility.Otod(res.getY()), 0, value_y[0], 0,
				res.size());
		int item_count = res.size();
		String[] message = new String[1];
		message[0] = "BM";
		line.setDataset(value_x, value_y, series_count, item_count, message);

		int memori_interval = 0;
		File one = new File(CUtility.PNGPATH);
		if(one.exists()==false)
			one.mkdirs();
		String pngfname = CUtility.PNGPATH + title + CUtility.getToDay();
		boolean isVisible = true;
		boolean isLegend = true;
		line.workChart(title, memori_interval, pngfname, isVisible, isLegend);
	}
	
	
	/**
	 * 時系列データの描画
	 * 
	 * @param value_x
	 *          value_x[i]= 時系列データ（描画の対象）x-値の配列
	 * @param value_y
	 *          value_y[i]= 時系列データ（描画の対象）y-値の配列
	 * 
	 * 制約　value_xとvalue_yは長さが一致することおよび各value_x[i]とvalue_y[i]も長さが一致すること。
	 * 
	 * @param messages　名称（各軸）長さはvalue_xと一致すること
	 * 
	 * @param title タイトル
	 *           
	 */
	public static void draw(double[][] value_x, double[][] value_y,String[] messages , String title) {
		
		TOChartLine line = new TOChartLine();
		
		assert(value_x.length == value_y.length);
		
		
		int series_count = value_x.length;
		
		int item_count = value_x[0].length;
		
		//double[][] value_x = new double[series_count][];

		//double[][] value_y = new double[series_count][];
		
		

		for(int i = 0; i< series_count ; i++){
		
		if(item_count != value_x[i].length)
			assert(false);
		

		if(item_count != value_y[i].length)
			assert(false);
	
	
		}
		
		line.setDataset(value_x, value_y, series_count, item_count, messages);

		int memori_interval = 0;
		
		File one = new File(CUtility.PNGPATH);
		if(one.exists()==false)
			one.mkdirs();
		String pngfname = CUtility.PNGPATH + title + CUtility.getToDay();
		
		boolean isVisible = true;
		boolean isLegend = true;
		
		line.workChart(title, memori_interval, pngfname, isVisible, isLegend);
	}

	
	/**
	 * 時系列データの描画
	 * 
	 * @param value_Axisx
	 *          value_Axisx[i]= xyz-軸のxー座標配列
	 * @param value_Axisy
	 *          value_Axisy[i]= xyz-軸のyー座標配列
	 * 
	 * 制約　value_xとvalue_yは長さが一致することおよび各value_x[i]とvalue_y[i]も長さが一致すること。
	 * 
	 * @param messages　名称（各軸）長さはvalue_xと一致すること
	 * 
	 * @param _x 描画対象データのxー座標
	 * 
	 * @param _y 描画対象データのxー座標
	 * 
	 * @param m 描画対象データのメッセージ
	 * 
	 * @param title タイトル
	 *           
	 */
	public static void draw(double[][] value_Axisx, double[][] value_Axisy,String[] messages , 
			double[] _x , double[] _y , String m,String title) {
		
		TOChartLine line = new TOChartLine();
		
		assert(value_Axisx.length == value_Axisy.length);
		
		
		int series_count = value_Axisx.length;
		
		int item_count = value_Axisx[0].length;
		
		

		for(int i = 0; i< series_count ; i++){
		
		if(item_count != value_Axisx[i].length)
			assert(false);
		

		if(item_count != value_Axisy[i].length)
			assert(false);
	
	
		}
		
		// データセットへ各データ（座標軸）および描画データを格納する。
		line.setDataset(value_Axisx, value_Axisy,  messages,
				_x,_y,m );

		int memori_interval = 0;
		
		File one = new File(CUtility.PNGPATH);
		if(one.exists()==false)
			one.mkdirs();
		String pngfname = CUtility.PNGPATH + title + CUtility.getToDay();
		
		boolean isVisible = true;
		boolean isLegend = true;
		
		line.workChart(title, memori_interval, pngfname, isVisible, isLegend);
	}

	
	
	/**
	 * 時系列データの描画
	 * 
	 * @param res 時系列データ（描画の対象）の配列
	 * @param messages 各時系列データの名前
	 * 	
	 * 制約　resとmessaesはサイズが一致していること。
	 * 
	 * @param title
	 *            タイトル
	 */
	public static void draw(CTS[] res, String[] messages , String title) {
		
		TOChartLine line = new TOChartLine();
		
		int series_count = res.length;
		
		int item_count = res[0].size();
		
		double[][] value_x = new double[series_count][];

		double[][] value_y = new double[series_count][];
		
		
		
		for(int i = 0; i< series_count ; i++){
		value_x[i] = new double[res[i].size()];
		// System.arraycopy(CUtility.itod(CUtility.Otoi(res.getX())), 0,
		// value_x[0], 0, res.size());
		
		System.arraycopy(CUtility.Otod(res[i].getX()), 0, value_x[i], 0,
				res[i].size());
		value_y[i] = new double[res[i].size()];
		
		System.arraycopy(CUtility.Otod(res[i].getY()), 0, value_y[i], 0,
				res[i].size());
		
		
		if(item_count != res[i].size())
			{
			
			System.out.println(i);
			
			System.out.println(i+","+item_count+","+res[i].size());
			
			assert(false);
			}
		
		
	
	
		}
		
		line.setDataset(value_x, value_y, series_count, item_count, messages);

		int memori_interval = 0;
		
		File one = new File(CUtility.PNGPATH);
		if(one.exists()==false)
			one.mkdirs();
		String pngfname = CUtility.PNGPATH + title + CUtility.getToDay();
		
		boolean isVisible = true;
		boolean isLegend = false;
		
		line.workChart(title, memori_interval, pngfname, isVisible, isLegend);
	}

	
	
	/**
	 * 時系列データの描画
	 * 
	 * @param res
	 *            時系列データ（描画の対象）
	 * @param res1
	 *            2番めの時系列データ
	 * 
	 * 制約　resとres1はx座標が一致していること。
	 * 
	 * @param title
	 * @param mes1
	 * @param mes2
	 *            タイトル
	 */
	public static void draw(CTS res, CTS res1, String title,String mes1,String mes2) {
		TOChartLine line = new TOChartLine();
		int series_count = 2;
		double[][] value_x = new double[series_count][];
		
		value_x[0] = new double[res.size()];
		value_x[1] = new double[res1.size()];
		// System.arraycopy(CUtility.itod(CUtility.Otoi(res.getX())), 0,
		// value_x[0], 0, res.size());
		System.arraycopy(CUtility.Otod(res.getX()), 0, value_x[0], 0,
				res.size());
		System.arraycopy(CUtility.Otod(res1.getX()), 0, value_x[1], 0,
				res1.size());
		
		double[][] value_y = new double[series_count][];
		
		value_y[0] = new double[res.size()];
		value_y[1] = new double[res1.size()];
		
		System.arraycopy(CUtility.Otod(res.getY()), 0, value_y[0], 0,
				res.size());
		System.arraycopy(CUtility.Otod(res1.getY()), 0, value_y[1], 0,
				res1.size());
		
		int item_count = res.size();
		
		String[] message = new String[2];
		message[0] = mes1;
		message[1] = mes2;
		
		line.setDataset(value_x, value_y, series_count, item_count, message);

		int memori_interval = 0;
		
		File one = new File(CUtility.PNGPATH);
		if(one.exists()==false)
			one.mkdirs();
		String pngfname = CUtility.PNGPATH + title + CUtility.getToDay();
		
		boolean isVisible = true;
		boolean isLegend = true;
		
		line.workChart(title, memori_interval, pngfname, isVisible, isLegend);
	}

	/**
	 * 時系列データの描画(plot)
	 * 
	 * @param res
	 *            時系列データ（描画の対象）
	 * @param title
	 *            タイトル
	 * @param x_title
	 *            X座標名称
	 * @param y_title
	 *            Y座標名称
	 */
	public static void drawPlot(CTS res, String title, String x_title,
			String y_title) {
		TOChartPlot plot = new TOChartPlot();
		int series_count = 1;
		double[][] value_x = new double[series_count][];
		value_x[0] = new double[res.size()];
		// System.arraycopy(CUtility.itod(CUtility.Otoi(res.getX())), 0,
		// value_x[0], 0, res.size());
		System.arraycopy(CUtility.Otod(res.getX()), 0, value_x[0], 0,
				res.size());
		double[][] value_y = new double[series_count][];
		value_y[0] = new double[res.size()];
		System.arraycopy(CUtility.Otod(res.getY()), 0, value_y[0], 0,
				res.size());
		int item_count = res.size();
		String[] message = new String[1];
		message[0] = "BM";
		plot.setDataset(value_x, value_y, series_count, item_count);

		String pngfname = title + CUtility.getToDay();

		plot.workChart(title, x_title, y_title, pngfname);
	}

	/**
	 * 　描画データの作成
	 * 
	 * @param lowwer
	 * @param upper
	 * @param integrand
	 * @param n
	 * @param isIntegrate
	 *            積分値を計算・表示する
	 * @param isDataPrint
	 *            データ値を印字する
	 * @return CTS
	 * 
	 */
	static public CTS getGrData(double lowwer, double upper,
			UnivariateFunction integrand, int n, boolean isIntegrate,boolean isDataPrint) {
		
		assert(lowwer < upper);
		

		double area = 0.;

		CTS res = new CTS();

		double dt = (upper - lowwer) / n;

		for (int i = 0; i < n; i++) {
			double x = dt * i + lowwer;
			double y = integrand.value(x);
			if (isIntegrate == true)
				area += dt * y;
			res.put(x, y);
			if(isDataPrint == true)
			logger.debug(x + "," + y);

		}

		if (isIntegrate == true)
			logger.debug("area[" + area + "]");
		return res;

	}

	/**
	 * 印字桁をそろえるための関数
	 */
	public static String getSp(int n) {
		String rtn = "";
		while (n-- > 0)
			rtn += " ";
		return rtn;
	}

	/**
	 * double型の小数点一位までの丸めとスケール（％表示のため）
	 * 
	 * 0.222222 -> 22.2 scale <--100 ika <--1
	 * 
	 * @param v
	 * @param scale
	 * @param ika
	 * 
	 * 
	 * @throws CAssertionException
	 * 
	 * @return formated value
	 * 
	 * 
	 */
	static public double cvDouble(double v, double scale, int ika)
			throws CAssertion {

		if (scale <= 0 && ika < 0) {

			String msg = "scale[" + scale + "] <= 0 && ika[" + ika + "] <= 0";

			throw new CAssertion(msg);
		}

		// スケール
		double drtn = v * scale;

		// 以下小数点
		double denominator = 1.0;

		// 割る値を計算
		while (ika-- > 0) {
			denominator *= 10.0;
		}

		// 掛ける
		drtn *= denominator;

		// 整数化する
		double rtn = (int) drtn;

		// 割る
		rtn /= denominator;

		// 戻す
		return rtn;

	}

	/**
	 * 試験
	 * @throws IOException 
	 * @throws NumberFormatException 
	 */
	public static void main(String[] arg) throws NumberFormatException, IOException {
		//for (int i = 0; i < 100; i++) {
		//	double x = Math.pow(2, i);
		//	System.out.println(i + ":[" + x + "][" + (int) x + "][" + (long) x
		//			+ "]");
		//}
		
		//CTS one = CUtility.fread2(arg[0]);
		//CUtility.draw(one, "alpha(0.7) percentile option price ");
		//CTS one = CUtility.fread(arg[0]);
		//CUtility.draw(one, "P(Xt>0,|1(-∞,0](Xs)ds > αt ");
		//String title = "CAlphaPercentileMonteCarlo sigma = 1 , nyu = 0.";
		
		CTS[] array = new CTS[3];
		
		CTS one = new CTS();
		
		array[0] = one;
		
		array[1] = one;
		
		array[2] = one;
		
		one.put(0., 1.0);
		one.put(1., 1.);
		
		String[] messages = new String[3];
		messages[0] = "x";
		messages[1] = "y";
		messages[2] = "z";
		
		
		String title = "test";
		CUtility.draw(array, messages,title);
		
	}

}
